--[[
Name: SpecialEvents-Aura-2.0
Revision: $Rev: 19932 $
Author: Tekkub Stoutwrithe (tekkub@gmail.com)
Website: http://www.wowace.com/
Description: Special events for Auras, (de)buffs gained, lost etc.
Dependencies: AceLibrary, AceEvent-2.0
--]]

local vmajor, vminor = "SpecialEvents-Aura-2.0", "$Revision: 19932 $"

if not AceLibrary then error(vmajor .. " requires AceLibrary.") end
if not AceLibrary:HasInstance("AceEvent-2.0") then error(vmajor .. " requires AceEvent-2.0.") end
if not AceLibrary:IsNewVersion(vmajor, vminor) then return end

local lib = {}
AceLibrary("AceEvent-2.0"):embed(lib)

local RL
----------------------------
--      mini-compost      --
----------------------------
local auraHeap = {}
setmetatable(auraHeap, { __mode = "kv" })

function lib:getheapsize()
	local i = 0
	for k,v in pairs(auraHeap) do
		i = i + 1
	end
	return i
end

local function acquire()
	local t = next(auraHeap)

	if t then
		auraHeap[t] = nil
		assert(not next(t), "A table in the compost heap has been modified!")
	else
		t = {}
	end

	return t
end

local function reclaim(t, d)
	if not t then return end

	if d and d > 0 then
		for i in pairs(t) do
			if type(t[i]) == "table" then
				reclaim(t[i], d - 1)
			end
			t[i] = nil
		end
	else
		-- if this isn't a deep reclaim, assume we're dealing with aura info
		t.index = nil
		t.name = nil
		t.rank = nil
		t.icon = nil
		t.count = nil
		t.dispellType = nil
	end

	auraHeap[t] = true
end

----------------------------
--     Initialization     --
----------------------------

local function registerevents(self)
	self:RegisterEvent("PLAYER_LEAVING_WORLD")
	self:RegisterEvent("UNIT_AURA", "AuraScan")
	self:RegisterEvent("UNIT_AURASTATE", "AuraScan")
	self:RegisterEvent("PLAYER_AURAS_CHANGED")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("PLAYER_FOCUS_CHANGED")
	-- check if RosterLib exists. We need to do that here and not earlier.
	RL = AceLibrary:HasInstance("RosterLib-2.0") and AceLibrary("RosterLib-2.0") or nil
	if RL then
		self:RegisterBucketEvent("RosterLib_UnitChanged",0.2)
	else
		self:RegisterEvent("PARTY_MEMBERS_CHANGED",1)
		self:RegisterEvent("RAID_ROSTER_UPDATE",1)
	end
end


-- Activate a new instance of this library
function activate(self, oldLib, oldDeactivate)
	if oldLib then
		self.vars = oldLib.vars
		if oldLib:IsEventRegistered("UNIT_AURA") then registerevents(self) end
		if oldLib:IsEventRegistered("RosterLib_UnitChanged") then
			oldLib:UnregisterAllEvents()
		end
	else
		self.vars = { buffs = {}, debuffs = {} }
	end

	if AceLibrary("AceEvent-2.0"):IsFullyInitialized() then
		self:PLAYER_ENTERING_WORLD()
	end

	self:RegisterEvent("PLAYER_ENTERING_WORLD")

	if oldDeactivate then oldDeactivate(oldLib) end
end


function lib:PLAYER_ENTERING_WORLD()
	self:ScanAllAuras()
	registerevents(self)
end


function lib:PLAYER_LEAVING_WORLD()
	self:UnregisterAllEvents()
	self:RegisterEvent("PLAYER_ENTERING_WORLD")
end


--------------------------------
--      Tracking methods      --
--------------------------------

function lib:PLAYER_AURAS_CHANGED()
	self:AuraScan("player")
	if GetNumRaidMembers() > 0 then
		local u
		for i=1,GetNumRaidMembers() do
			if UnitIsUnit("raid"..i, "player") then u = "raid"..i end
		end
		self:AuraScan(u)
	end
end


function lib:PLAYER_TARGET_CHANGED()
	self:AuraScan("target")
	self:TriggerEvent("SpecialEvents_AuraTargetChanged")
end


function lib:PLAYER_FOCUS_CHANGED()
	self:AuraScan("focus")
	self:TriggerEvent("SpecialEvents_AuraFocusChanged")
end


function lib:RosterLib_UnitChanged(units)
	for unit in pairs(units) do
		if unit and UnitExists(unit) then
			self:AuraScan(unit)
		end
	end
	if GetNumRaidMembers() > 0 then
		self:TriggerEvent("SpecialEvents_AuraRaidRosterUpdate")
	else
		self:TriggerEvent("SpecialEvents_AuraPartyMembersChanged")
	end
end


function lib:PARTY_MEMBERS_CHANGED()
	if UnitExists("pet") then self:AuraScan("pet") end

	for i=1,4 do
		if UnitExists("party"..i) then self:AuraScan("party"..i) end
		if UnitExists("partypet"..i) then self:AuraScan("partypet"..i) end
	end
	self:TriggerEvent("SpecialEvents_AuraPartyMembersChanged")
end


function lib:RAID_ROSTER_UPDATE()
	for i=1,40 do
		if UnitExists("raid"..i) then self:AuraScan("raid"..i) end
		if UnitExists("raidpet"..i) then self:AuraScan("raidpet"..i) end
	end
	self:TriggerEvent("SpecialEvents_AuraRaidRosterUpdate")
end

function lib:ScanAllAuras()
	if UnitExists("player") then self:AuraScan("player") end
	if UnitExists("pet") then self:AuraScan("pet") end

	for i=1,4 do
		if UnitExists("party"..i) then self:AuraScan("party"..i) end
		if UnitExists("partypet"..i) then self:AuraScan("partypet"..i) end
	end

	for i=1,40 do
		if UnitExists("raid"..i) then self:AuraScan("raid"..i) end
		if UnitExists("raidpet"..i) then self:AuraScan("raidpet"..i) end
	end

	if UnitExists("target") then self:AuraScan("target") end
	if UnitExists("focus") then self:AuraScan("focus") end
--~~ 	if UnitExists("mouseover") then self:AuraScan("mouseover") end
end


function lib:AuraScan(targ)
	local maxbuffs, maxdebuffs, unit = 32, 40, targ or arg1
	local oldd, oldb = self.vars.debuffs[unit], self.vars.buffs[unit]
	local newd, newb = acquire(), acquire()

	for index=1,maxbuffs do
		local name, rank, icon, count

		-- Honest, there is no guaranteed correlation between the
		-- result of GetPlayerBuff and the index used by UnitBuff.
		-- GetPlayerBuff sucks, but we need it for backwards
		-- compatability.
		if unit == "player" then
			local pindex = GetPlayerBuff(index, "HELPFUL")
			name, rank = GetPlayerBuffName(pindex)
			icon = GetPlayerBuffTexture(pindex)
			count = GetPlayerBuffApplications(pindex)
		else
			name, rank, icon, count = UnitBuff(unit, index)
		end
		
		if name then
			local buffIndex = string.format("%s_%s_%s", name, rank, icon)
			local t = acquire()
			t.index  = index
			t.name   = name
			t.rank   = rank
			t.icon   = icon
			t.count  = count

			newb[buffIndex] = t
		end
	end
	
	for index=1,maxdebuffs do
		local name, rank, icon, count, dispellType

		if unit == "player" then
			local pindex = GetPlayerBuff(index, "HARMFUL")
			name, rank = GetPlayerBuffName(pindex)
			icon = GetPlayerBuffTexture(pindex)
			count = GetPlayerBuffApplications(pindex)
			dispellType = GetPlayerBuffDispelType(pindex)
		else
			name, rank, icon, count, dispellType = UnitDebuff(unit, index)
		end



		if name and index > 0 then
			local debuffIndex = string.format("%s_%s_%s", name, rank, icon)
			local t = acquire()
			t.index  = index
			t.name   = name
			t.rank   = rank
			t.icon   = icon
			t.count  = count
			t.dispellType = dispellType or ""

			newd[debuffIndex] = t
		end
	end

	self.vars.buffs[unit] = newb
	self.vars.debuffs[unit] = newd

	if oldb then
		for buffIndex,buffInfo in pairs(oldb) do
			if type(buffInfo) ~= "table" then
				local t = ""
				for d,i in pairs(oldb) do t = t.." "..d.."="..i end
				error(string.format("Buff error! Unit: %s, Table dump:%s", unit, t))

			elseif not newb[buffIndex] then
				-- the ';' shows where the previos SEEA-2.0 API ended
				-- unit, name; count, icon, rank
				self:TriggerEvent("SpecialEvents_UnitBuffLost", unit,
					buffInfo.name, buffInfo.count, buffInfo.icon, buffInfo.rank)

				if unit == "player" then
					-- name; count, icon rank
					self:TriggerEvent("SpecialEvents_PlayerBuffLost",
						buffInfo.name, buffInfo.count, buffInfo.icon, buffInfo.rank)
				end
			end
		end
	end

	if oldd then
		for debuffIndex,debuffInfo in pairs(oldd) do
			if type(debuffInfo) ~= "table" then
				local t = ""
				for d,i in pairs(oldd) do t = t.." "..d.."="..i end
				error(string.format("Debuff error! Unit: %s, Table dump:%s", unit, t))

			elseif not newd[debuffIndex] then
				-- name; count, dispellType, icon, rank
				self:TriggerEvent("SpecialEvents_UnitDebuffLost", unit,
					debuffInfo.name, debuffInfo.count, debuffInfo.dispellType, debuffInfo.icon, debuffInfo.rank)

				if unit == "player" then
					-- name; count, dispellType, icon, rank
					self:TriggerEvent("SpecialEvents_PlayerDebuffLost",
						debuffInfo.name, debuffInfo.count, debuffInfo.dispellType, debuffInfo.icon, debuffInfo.rank)
				end
			end
		end
	end

	for buffIndex,buffInfo in pairs(newb) do
		assert(type(buffInfo) == "table",
			string.format("Buff: %s - Value not a table: %s",
				buffIndex, type(buffInfo) == "table" and "table" or buffInfo or "nil"))

		if (not oldb or not oldb[buffIndex]) then
			-- unit, name; index, count, icon, rank
			self:TriggerEvent("SpecialEvents_UnitBuffGained", unit,
				buffInfo.name, buffInfo.index, buffInfo.count, buffInfo.icon, buffInfo.rank)

			if unit == "player" then
				-- name, index; count, icon, rank
				self:TriggerEvent("SpecialEvents_PlayerBuffGained",
						  buffInfo.name, buffInfo.index, buffInfo.count, buffInfo.icon, buffInfo.rank)
			end

		elseif oldb and oldb[buffIndex] and oldb[buffIndex].count ~= buffInfo.count then
			-- unit, name, index, count, icon, rank
			self:TriggerEvent("SpecialEvents_UnitBuffCountChanged", unit,
				buffInfo.name, buffInfo.index, buffInfo.count, buffInfo.icon, buffInfo.rank)

			if unit == "player" then
				-- name, index, count, icon, rank
				self:TriggerEvent("SpecialEvents_PlayerBuffCountChanged",
						  buffInfo.name, buffInfo.index, buffInfo.count, buffInfo.icon, buffInfo.rank)
			end
		end
	end

	for debuffIndex,debuffInfo in pairs(newd) do
		assert(type(debuffInfo) == "table",
			string.format("Debuff: %s - Value not a table: %s",
				debuffIndex, type(debuffInfo) == "table" and "table" or debuffInfo or "nil"))

		if (not oldd or not oldd[debuffIndex]) then
			-- unit, name, count, dispellType, icon; rank, index
			self:TriggerEvent("SpecialEvents_UnitDebuffGained", unit,
				debuffInfo.name, debuffInfo.count, debuffInfo.dispellType, debuffInfo.icon, debuffInfo.rank, debuffInfo.index)

			if unit == "player" then
				-- name, count, dispellType, icon; rank, index
				self:TriggerEvent("SpecialEvents_PlayerDebuffGained",
						  debuffInfo.name, debuffInfo.count, debuffInfo.dispellType, debuffInfo.icon, debuffInfo.rank, debuffInfo.index)
			end

		elseif oldd and oldd[debuffIndex] and oldd[debuffIndex].count ~= debuffInfo.count then
			-- unit, name, count, dispellType, icon; rank, index
			self:TriggerEvent("SpecialEvents_UnitDebuffCountChanged", unit,
				debuffInfo.name, debuffInfo.count, debuffInfo.dispellType, debuffInfo.icon, debuffInfo.rank, debuffInfo.index)
			
			if unit == "player" then
				-- name, count, dispellType, icon; rank, index
				self:TriggerEvent("SpecialEvents_PlayerDebuffCountChanged",
						  debuffInfo.name, debuffInfo.count, debuffInfo.dispellType, debuffInfo.icon, debuffInfo.rank, debuffInfo.index)
			end
		end
	end

	if oldb then reclaim(oldb, 1) end
	if oldd then reclaim(oldd, 1) end
end


-----------------------------
--      Query Methods      --
-----------------------------

function lib:UnitHasBuff(unit, name, rank, icon)
	if not self.vars.buffs[unit] then return end

	local unitBuffs = self.vars.buffs[unit]

	if name and rank and icon then
		local buffInfo = unitBuffs[string.format("%s_%s_%s", name, rank, icon)]
		if buffInfo then
			-- index; count, icon, rank
			return buffInfo.index, buffInfo.count, buffInfo.icon, buffInfo.rank
			-- there should probably be an 'else return nil' here
		end
	end

	for buffIndex,buffInfo in pairs(unitBuffs) do
		if (not name or name == buffInfo.name) and
			(not rank or rank == buffInfo.rank) and
			(not icon or icon == buffInfo.icon) then

			-- index; count, icon, rank
			return buffInfo.index, buffInfo.count, buffInfo.icon, buffInfo.rank
		end
	end
end


function lib:UnitHasDebuff(unit, name, rank, icon)
	if not self.vars.debuffs[unit] then return end

	local unitDebuffs = self.vars.debuffs[unit]

	if name and rank and icon then
		local debuffInfo = unitDebuffs[string.format("%s_%s_%s", name, rank, icon)]
		if debuffInfo then
			-- index, count; icon, rank
			return debuffInfo.index, debuffInfo.count, debuffInfo.icon, debuffInfo.rank
		end
	end

	for debuffIndex,debuffInfo in pairs(unitDebuffs) do
		if (not name or name == debuffInfo.name) and
			(not rank or rank == debuffInfo.rank) and
			(not icon or icon == debuffInfo.icon) then

			-- index, count; icon, rank
			return debuffInfo.index, debuffInfo.count, debuffInfo.icon, debuffInfo.rank
		end
	end
end


function lib:UnitHasDebuffType(unit, dispellType)
	if not self.vars.debuffs[unit] then return end

	local unitDebuffs = self.vars.debuffs[unit]

	for debuffIndex,debuffInfo in pairs(unitDebuffs) do
		if debuffInfo.dispellType == dispellType then
			-- index, count, icon, rank
			return debuffInfo.index, debuffInfo.count, debuffInfo.icon, debuffInfo.rank
		end
	end
end


function lib:BuffIter(unit)
	local unitBuffs = self.vars.buffs[unit]
	local i, buffInfo

	return function()
		if not unitBuffs then return end

		i, buffInfo = next(unitBuffs, i)
		if buffInfo then
			-- name, index; count, icon, rank
			return buffInfo.name, buffInfo.index, buffInfo.count, buffInfo.icon, buffInfo.rank
		end
	end
end

function lib:DebuffIter(unit)
	local unitDebuffs = self.vars.debuffs[unit]
	local i, debuffInfo

	return function()
		if not unitDebuffs then return end

		i, debuffInfo = next(unitDebuffs, i)
		if debuffInfo then
			-- name, count, dispellType, icon; rank, index
			return debuffInfo.name, debuffInfo.count, debuffInfo.dispellType, debuffInfo.icon, debuffInfo.rank, debuffInfo.index
		end
	end
end

-------------------------------
--  Say hello to AceLibrary  --
-------------------------------
AceLibrary:Register(lib, vmajor, vminor, activate)

